""" Pspring Mapping decorator 
"""
import logging
import json
import inspect
import requests

from requests.exceptions import HTTPError

from pspring import *


logger = logging.getLogger("pspring-rest-client")

class Mapping():
    """A decorator for methods inside a class that is decorated by RestClient. When this decorator is applied to a method, turns that method to
        send a http request based on the arguments

        Example:
        ```python
        
        import pspringrestclient

        @RestClient(url="https://myapi.com")
        class TestClient():
            @Mapping(url="/users",method="GET")
            def get_users(self):
                pass

        client = TestClient()
        client.get_users() # will trigger a http GET requedt to https://myapi.com/users

        ```
        The list of features of the mapping decorator are
        
        * provides easy access to make methods of a class that send http requests
        * kargs that are recognized are, method, url, data, timeout
        * can have dynamic variables in the url which will be evaluated with the function arguments
        * send non json payload in POST
        * set proxies in the request
        * set query string in HTTP request
        * send json payload in POST
    """
    def __init__(self,*args,**kargs):
        self.method = kargs.get("method")
        self.url = kargs.get("url")
        self.data = kargs.get("data")
        self.timeout = kargs.get("timeout")

    def __call__(self,func_obj):
        def new_func(*args,**kargs):
            argspec = inspect.getfullargspec(func_obj)
            argument_names = argspec[0]
            url = self.url
            for i in range(len(argument_names)):
                if(len(args) > i ):
                    url = url.replace("{"+argument_names[i]+"}",str(args[i]))

            for (karg_key,karg_val) in kargs.items():
                url = url.replace("{"+karg_key+"}",str(karg_val))

            self_obj = args[0]
            kargs.update({
                "url" : self_obj.getUrl()+url
            })
            kargs.update({
                "method" : self.method
            })

            if(self.timeout != None):
                kargs.update({
                    "timeout" : float(self.timeout)
                })

            if self.data != None:
                kargs.update({
                    "data" : self.data
                })

            kargs_to_update = func_obj(*args,**kargs)

            if(kargs != None and kargs_to_update != None):
                kargs.update(kargs_to_update)

            if(hasattr(self_obj,"queryString") and self_obj.queryString != None):
                kargs.update({
                    "url" : kargs.get("url") + "?" + self_obj.queryString
                })

            if (hasattr(self_obj,"proxies") and self_obj.proxies != None):
                kargs.update({
                    "proxies" : self_obj.proxies
                })

            if(hasattr(self_obj,"timeout") and self_obj.timeout != None):
                kargs.update({
                    "timeout" : float(self_obj.timeout)
                })

            if (hasattr(self_obj,"data") and self_obj.data != None):
                kargs.update({
                    "data" : self_obj.data
                })

            if (hasattr(self_obj,"json") and self_obj.json != None):
                kargs.update({
                    "json" : self_obj.json
                })


            return self_obj.send(**kargs)
        return new_func



import sys
sys.path.append("./deps")
sys.path.append(".")
import os
os.environ["pspring.aws.secretsMngr.secretName"] = "sales-ctp-prod"
os.environ["pspring.soo.security.cacheTable"] = "dev-token-cache"

from pspring import *
from unittest import mock
import logging
import logging.config
from pythoncloudlogger import *
from loggingconfig import config as filelogconfig
logging.config.dictConfig(filelogconfig)
logger = logging.getLogger(__name__)
import requests
from pspringrestclient import RestClient,Mapping

from mock_response import MockResponse

def test_mapping_url_method():
    with mock.patch("requests.request") as mock_get:
        mock_get.return_value = MockResponse()

        @RestClient(url="https://reqres.in")
        class Test():
            @Mapping(url="/api/users?page=2",method="GET")
            def get_users(self,*args,**kargs):
                pass

        test_client = Test()
        response = test_client.get_users()
        assert response.get("body").get("message") == "Hello from the other side"

def test_mapping_dynamic_url():
    with mock.patch("requests.request") as mock_get:
        mock_get.return_value = MockResponse()

        @RestClient(url="https://reqres.in")
        class Test():
            @Mapping(url="/api/users?page={page_number}",method="GET")
            def get_users(self,page_number,*args,**kargs):
                self._url = kargs.get("url")
        
        test_client = Test()
        response = test_client.get_users(1)
        assert test_client._url == "https://reqres.in/api/users?page=1"

def test_mapping_post_json():
    with mock.patch("requests.request") as mock_get:
        mock_get.return_value = MockResponse()

        @RestClient(url="https://reqres.in")
        class Test():
            @Mapping(url="/api/users",method="POST")
            def create_user(self,*args,**kargs):
                self.json = {"message":"hi"}
        
        test_client = Test()
        response = test_client.create_user(1)
        assert response.get("body").get("message") == "Hello from the other side"

def test_mapping_post_xml():
    with mock.patch("requests.request") as mock_get:
        mock_get.return_value = MockResponse(content_type="xml")

        @RestClient(url="https://reqres.in")
        class Test():
            @Mapping(url="/api/users",method="POST")
            def create_user(self,*args,**kargs):
                self.data = "<xml><user></user></xml>"
        
        test_client = Test()
        test_client.addHeader("Content-Type","text/xml")
        response = test_client.create_user(1)
        assert response.get("body") == "<xml><message>Hello from the other side</message></xml>"

def test_mapping_get():
    with mock.patch("requests.request") as mock_get:
        mock_get.return_value = MockResponse()

        @RestClient(url="https://reqres.in")
        class Test():
            @Mapping(url="/api/users?page=2",method="GET")
            def get_users(self,*args,**kargs):
                pass
        logger.info("Test here")    
        def timestamp_middleware(request,response):
            if response is None:
                headers = request.get("headers")
                headers.update({
                    "timestamp1" : "121342423423"
                })
                logger.info({
                    "message":"in timestamp middleware",
                    "request":request
                })
            else:
                logger.info({
                    "message":"in timestamp middleware",
                    "response":response
                })
        RestClient.middlewares.append(timestamp_middleware)

        testClient = Test()
        response = testClient.get_users()
        logger.info(response)
        assert response.get("body",None) != None

def test_mapping_query_string():
    """test query string set manually
    """
    with mock.patch("requests.request") as mock_get:
        mock_get.return_value = MockResponse()

        @RestClient(url="https://reqres.in")
        class Test():
            @Mapping(url="/api/users",method="GET")
            def get_users(self,page_number,*args,**kargs):
                self.queryString = "page=4"
        
        test_client = Test()
        def send_decorator(func_obj):
            def new_send(*args,**kargs):
                self = test_client
                self.kargs = kargs
                func_obj(*args,**kargs)
            return new_send
        test_client.send = send_decorator(test_client.send)
        response = test_client.get_users(1)
        assert test_client.kargs.get("url") == "https://reqres.in/api/users?page=4"


""" Test suite for RestClient class
"""
import os
import sys
from mock_response import MockResponse
from requests.exceptions import HTTPError
import requests
import logging.config
import logging
from unittest import mock
from loggingconfig import config as filelogconfig

sys.path.append("./deps")
sys.path.append(".")

from pspringrestclient import RestClient, Mapping
from pythoncloudlogger import *
from pspring import *

logging.config.dictConfig(filelogconfig)
logger = logging.getLogger(__name__)


def test_restclient_add_header():
    """Test for addHeader method
    """
    @RestClient(url="https://reqres.in")
    class Test():
        """ Test client
        """
        pass

    test_client = Test()
    test_client.addHeader("key", "value")
    assert test_client.headers.get("key") == "value"


def test_restclient_add_middleware():
    """Test for add_middleware method
    """
    with mock.patch("requests.request") as mock_get:
        mock_get.return_value = MockResponse()

        @RestClient(url="https://reqres.in")
        class Test():
            """ Test client
            """
            pass

        def add_timestamp_middleware(request, response):
            headers = request.get("headers")
            headers["timestamp"] = "12345"
        test_client = Test()
        test_client.add_middleware(add_timestamp_middleware)
        response = test_client.send(url="https://abc.com/users", method="GET")
        assert test_client.headers.get("timestamp") == "12345"

def test_restclient_class_middleware():
    """Test for class level middlewares
    """
    with mock.patch("requests.request") as mock_get:
        mock_get.return_value = MockResponse()

        @RestClient(url="https://reqres.in")
        class Test():
            middlewares=[]
            """ Test client
            """
            pass

        def add_timestamp_middleware(request, response):
            headers = request.get("headers")
            headers["timestamp"] = "12345"

        Test.middlewares.append(add_timestamp_middleware)
        test_client = Test()
        response = test_client.send(url="https://abc.com/users", method="GET")
        assert test_client.headers.get("timestamp") == "12345"


def test_restclient_finalize():
    """Test for finalize method
    """
    with mock.patch("requests.request") as mock_get:
        mock_get.return_value = MockResponse()

        @RestClient(url="https://reqres.in")
        class Test():
            """TestClient
            """
            def finalize(self):
                self.headers["trackingId"] = "3456"

        test_client = Test()
        response = test_client.send(url="https://abc.com/users", method="GET")
        assert test_client.headers.get("trackingId") == "3456"


def test_restclient_json_content_type():
    """Test for json content type
    """
    with mock.patch("requests.request") as mock_get:
        mock_get.return_value = MockResponse()

        @RestClient(url="https://reqres.in")
        class Test():
            """ Test Client
            """
            pass

        test_client = Test()
        response = test_client.send(url="https://abc.com/users", method="GET")
        assert response.get("body").get(
            "message") == "Hello from the other side"


def test_restclient_xml_content_type():
    """Test for xml content type
    """
    with mock.patch("requests.request") as mock_get:
        mock_get.return_value = MockResponse(content_type="xml")

        @RestClient(url="https://reqres.in")
        class Test():
            """ Test client
            """
            pass

        test_client = Test()
        response = test_client.send(url="https://abc.com/users", method="GET")
        assert response.get(
            "body") == "<xml><message>Hello from the other side</message></xml>"


def test_restclient_500_error():
    """Test for 500 error
    """
    with mock.patch("requests.request") as mock_get:
        mock_get.return_value = MockResponse(error_code=500)

        @RestClient(url="https://reqres.in")
        class Test():
            """ Test client
            """
            pass

        test_client = Test()
        exception = False
        try:
            response = test_client.send(
                url="https://abc.com/users", method="GET")
        except HTTPError as e:
            exception = True
        assert exception == True


def test_restclient_geturl():
    """Test for getUrl method
    """
    with mock.patch("requests.request") as mock_get:
        mock_get.return_value = MockResponse(error_code=500)

        @RestClient(url="https://reqres.in")
        class Test():
            """ Test client
            """
            pass

        test_client = Test()
        assert test_client.getUrl() == "https://reqres.in"



""" Pspring Rest Client module provides an easy interface and decorators to turn any class into a http client
"""
import logging
import json
import inspect
import requests

from requests.exceptions import HTTPError

from pspring import *


logger = logging.getLogger("pspring-rest-client")

class PayloadException(Exception):
    """[summary]

    Args:
        Exception ([type]): [description]
    """
    def __init__(self,*args):
        super().__init__(*args)
        self.response = args[2]
        self.statusCode = args[1]

class RestClient():
    """ RestClient is a class decorator that adds below capabilities to the class
    
        * Adds a method called add_header with which we can add headers to the http request
        * Adds a method called add_middleware, this method will allow hooks to modify the http request and response before it is returned
        * Adds a method called send, which will trigger the http request
        * Adds a method called finalize, which will be a pre-hook before http request is sent
        * Adds a method called getUrl which retuns the current url
    """
    middlewares = []
    def __init__(self,*args,**kargs):
        self.url = kargs.get("url")
        self.headers = kargs.get("headers",{})
        self.middlewares = kargs.get("middlewares",[])
        self.timeout = kargs.get("timeout")
        self.responsemapper = kargs.get("responsemapper")

    def __getattribute__(self, name):
        """
        Called for every attribute access in this class.
        Allows callback functions to be passed as values for instance variables by replacing the instance variable
        access functionality with an additional check for whether the instance variable is a callback, calling it and
        returning its result if it is.
        """
        attribute_value = object.__getattribute__(self, name)  # Existing behavior
        if name not in type(self).__dict__.keys() and callable(attribute_value):  # First condition ensures that class methods are never called here.
            return attribute_value()
        else:
            return attribute_value
            
    def __call__(self,class_obj):
        prev_init = class_obj.__init__

        def constructor(*args,**kargs):

            self_orig = args[0]
            def middleware(*args,**kargs):
                index = args[0] if len(args) == 1 else None
                def newfunc(func_obj):
                    self_orig.add_middleware(func_obj,index)
                    return func_obj
                return newfunc
            self_orig.headers = self.headers
            self_orig.url = self.url
            self_orig.middleware = middleware
            if(hasattr(self_orig,"middlewares") and self_orig.middlewares is not None):
                self_orig.middlewares = self_orig.middlewares + RestClient.middlewares+ self.middlewares
            else:
                self_orig.middlewares = RestClient.middlewares+ self.middlewares
            prev_init(*args,**kargs)

        def add_middleware(self_orig,func_obj,index=None):
            if index is not None:
                self_orig.middlewares.insert(index,func_obj)
            else:
                self_orig.middlewares.append(func_obj)

        def addHeader(self_orig,name,value):
            self_orig.headers.update({name:value})

        def clearHeader(self_orig):
            self_orig.headers = {}

        def send(*args,**kargs):
            self_orig = args[0]
            additional_args = self_orig.finalize()

            if isinstance(additional_args,dict):
                kargs.update(additional_args)

            if kargs.get("timeout") == None and self.timeout != None:
                kargs["timeout"] = float(self.timeout)

            kargs["headers"] = self_orig.headers

            for middleware in self_orig.middlewares:
                middleware(kargs,None)
            
            logger.info({
                "message" : "request details",
                "timeout" : kargs.get("timeout"),
                "method" : kargs.get("method"),
                "url" : kargs.get("url"),
                "data" : kargs.get("data"),
                "json" : kargs.get("json"),
                "proxies" : kargs.get("proxies"),
                "headers" : self_orig.headers
            })
            try:
                response = requests.request(**kargs)

                finalresponse = {}
                response.raise_for_status()

                if "json" in response.headers.get("Content-Type",""):
                    response_json = {}
                    if response.status_code != 204:
                        response_json = response.json()

                    if self.responsemapper != None:
                        response_json = self.responsemapper.map(response_json)
                    
                    finalresponse = {
                        "body":response_json,
                        "headers" : response.headers
                    }

                    logger.info({
                        "message" : "response details",
                        "method" : kargs.get("method"),
                        "url" : kargs.get("url"),
                        "data" : kargs.get("data"),
                        "json" : kargs.get("json"),
                        "proxies" : kargs.get("proxies"),
                        "headers" : self_orig.headers,
                        "status_code" : response.status_code,
                        "responseHeaders" : str(response.headers),
                        "response" : response_json,
                        "elapsed" : response.elapsed.total_seconds()
                    })
                
                else:
                    
                    finalresponse = {
                        "body":response.text,
                        "headers" : response.headers
                    }

                    logger.info({
                        "message" : "response details",
                        "method" : kargs.get("method"),
                        "url" : kargs.get("url"),
                        "data" : kargs.get("data"),
                        "json" : kargs.get("json"),
                        "proxies" : kargs.get("proxies"),
                        "headers" : self_orig.headers,
                        "status_code" : response.status_code,
                        "responseHeaders" : str(response.headers),
                        "response" : response.text,
                        "elapsed" : response.elapsed.total_seconds()
                    })

                for middleware in reversed(self_orig.middlewares):
                    middleware(kargs,finalresponse)
                return finalresponse
            except HTTPError as ex:
                logger.error({
                    "message" : str(ex)
                })
                logger.info({
                    "message" : "response details",
                    "method" : kargs.get("method"),
                    "url" : kargs.get("url"),
                    "data" : kargs.get("data"),
                    "proxies" : kargs.get("proxies"),
                    "headers" : self_orig.headers,
                    "status_code" : response.status_code,
                    "responseHeaders" : response.headers,
                    "response" : response.text,
                    "elapsed" : response.elapsed.total_seconds()
                })

                if hasattr(self_orig,"get_error_code"):
                    try:
                        if "json" in response.headers.get("Content-Type",""):
                            error_code = self_orig.get_error_code(ex.response.json())
                            ex.code = error_code
                    except Exception as excep_code:
                        excep_code_str = str(excep_code)
                        logger.error({
                            "message" : f"Exception while retrieving error code : {excep_code_str}"
                        })
                raise ex



        def getUrl(*args,**kargs):
            self_orig = args[0]
            return self_orig.url

        def finalize(*args,**kargs):
            pass

        class_obj.__init__ = constructor
        class_obj.addHeader = addHeader
        class_obj.clearHeader = clearHeader
        class_obj.add_middleware = add_middleware
        class_obj.send = send
        class_obj.getUrl = getUrl
        if not hasattr(class_obj,"finalize"):
            class_obj.finalize = finalize

        return class_obj
